@@ -52,18 +52,13 @@ class Gencontrol(Base):
             'headers-all': config.SchemaItemBoolean(),
             'installer': config.SchemaItemBoolean(),
             'libc-dev': config.SchemaItemBoolean(),
+            'meta': config.SchemaItemBoolean(),
             'tools-unversioned': config.SchemaItemBoolean(),
             'tools-versioned': config.SchemaItemBoolean(),
             'source': config.SchemaItemBoolean(),
         }
     }
 
-    env_flags = [
-        ('DEBIAN_KERNEL_DISABLE_DEBUG', 'disable_debug', 'debug infos'),
-        ('DEBIAN_KERNEL_DISABLE_INSTALLER', 'disable_installer', 'installer modules'),
-        ('DEBIAN_KERNEL_DISABLE_SIGNED', 'disable_signed', 'signed code'),
-    ]
-
     def __init__(self, config_dirs=["debian/config"],
                  template_dirs=["debian/templates"]):
         super(Gencontrol, self).__init__(
@@ -73,26 +68,11 @@ class Gencontrol(Base):
         self.process_changelog()
         self.config_dirs = config_dirs
 
-        for env, attr, desc in self.env_flags:
-            setattr(self, attr, False)
-            if os.getenv(env):
-                if self.changelog[0].distribution == 'UNRELEASED':
-                    import warnings
-                    warnings.warn(f'Disable {desc} on request ({env} set)')
-                    setattr(self, attr, True)
-                else:
-                    raise RuntimeError(
-                        f'Unable to disable {desc} in release build ({env} set)')
-
     def _setup_makeflags(self, names, makeflags, data):
         for src, dst, optional in names:
             if src in data or not optional:
                 makeflags[dst] = data[src]
 
-    def _substitute_file(self, template, vars, target, append=False):
-        with open(target, 'a' if append else 'w') as f:
-            f.write(self.substitute(self.templates[template], vars))
-
     def do_main_setup(self, vars, makeflags, extra):
         super(Gencontrol, self).do_main_setup(vars, makeflags, extra)
         makeflags.update({
@@ -101,7 +81,8 @@ class Gencontrol(Base):
             'ABINAME': self.abiname_version + self.abiname_part,
             'SOURCEVERSION': self.version.complete,
         })
-        makeflags['SOURCE_BASENAME'] = self.vars['source_basename']
+        makeflags['SOURCE_BASENAME'] = vars['source_basename']
+        makeflags['SOURCE_SUFFIX'] = vars['source_suffix']
 
         # Prepare to generate debian/tests/control
         self.tests_control = self.process_packages(
@@ -110,7 +91,16 @@ class Gencontrol(Base):
 
         self.installer_packages = {}
 
-        if not self.disable_installer and self.config.merge('packages').get('installer', True):
+        if os.getenv('DEBIAN_KERNEL_DISABLE_INSTALLER'):
+            if self.changelog[0].distribution == 'UNRELEASED':
+                import warnings
+                warnings.warn('Disable installer modules on request '
+                              '(DEBIAN_KERNEL_DISABLE_INSTALLER set)')
+            else:
+                raise RuntimeError(
+                    'Unable to disable installer modules in release build '
+                    '(DEBIAN_KERNEL_DISABLE_INSTALLER set)')
+        elif self.config.merge('packages').get('installer', True):
             # Add udebs using kernel-wedge
             kw_env = os.environ.copy()
             kw_env['KW_DEFCONFIG_DIR'] = 'debian/installer'
@@ -140,11 +130,8 @@ class Gencontrol(Base):
             # configuration errors before building linux-signed.
             build_signed = {}
             for arch in arches:
-                if not self.disable_signed:
-                    build_signed[arch] = self.config.merge('build', arch) \
-                                                    .get('signed-code', False)
-                else:
-                    build_signed[arch] = False
+                build_signed[arch] = self.config.merge('build', arch) \
+                                                .get('signed-code', False)
 
             for package in udeb_packages:
                 # kernel-wedge currently chokes on Build-Profiles so add it now
@@ -181,23 +168,44 @@ class Gencontrol(Base):
 
     def do_main_packages(self, packages, vars, makeflags, extra):
         packages.extend(self.process_packages(
-            self.templates["control.main"], self.vars))
+            self.templates["control.main"], vars))
+
+        # Only build the metapackages if their names won't exactly match
+        # the packages they depend on
+        do_meta = self.config.merge('packages').get('meta', True) \
+            and vars['source_suffix'] != '-' + vars['version']
+
         if self.config.merge('packages').get('docs', True):
             packages.extend(self.process_packages(
-                self.templates["control.docs"], self.vars))
+                self.templates["control.docs"], vars))
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.docs.meta"], vars))
+                self.substitute_debhelper_config(
+                    'docs.meta', vars,
+                    '%(source_basename)s-doc%(source_suffix)s' % vars)
         if self.config.merge('packages').get('tools-unversioned', True):
             packages.extend(self.process_packages(
-                self.templates["control.tools-unversioned"], self.vars))
+                self.templates["control.tools-unversioned"], vars))
         if self.config.merge('packages').get('tools-versioned', True):
             packages.extend(self.process_packages(
-                self.templates["control.tools-versioned"], self.vars))
+                self.templates["control.tools-versioned"], vars))
+            self.substitute_debhelper_config('perf', vars,
+                                              'linux-perf-%(version)s' % vars)
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.tools-versioned.meta"], vars))
+                self.substitute_debhelper_config('perf.meta', vars,
+                                                  'linux-perf')
         if self.config.merge('packages').get('source', True):
             packages.extend(self.process_packages(
-                self.templates["control.sourcebin"], self.vars))
-
-        self._substitute_file('perf.lintian-overrides', self.vars,
-                              'debian/linux-perf-%s.lintian-overrides' %
-                              self.vars['version'])
+                self.templates["control.sourcebin"], vars))
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.sourcebin.meta"], vars))
+                self.substitute_debhelper_config(
+                    'sourcebin.meta', vars,
+                    '%(source_basename)s-source%(source_suffix)s' % vars)
 
     def do_indep_featureset_setup(self, vars, makeflags, featureset, extra):
         makeflags['LOCALVERSION'] = vars['localversion']
@@ -257,11 +265,8 @@ class Gencontrol(Base):
             makeflags['ABINAME'] = vars['abiname'] = \
                 self.abiname_version + abiname_part
 
-        if not self.disable_signed:
-            build_signed = self.config.merge('build', arch) \
-                                      .get('signed-code', False)
-        else:
-            build_signed = False
+        build_signed = self.config.merge('build', arch) \
+                                  .get('signed-code', False)
 
         # Some userland architectures require kernels from another
         # (Debian) architecture, e.g. x32/amd64.
@@ -336,8 +341,9 @@ class Gencontrol(Base):
         ('compiler', 'COMPILER', False),
         ('compiler-filename', 'COMPILER', True),
         ('kernel-arch', 'KERNEL_ARCH', False),
-        ('cflags', 'CFLAGS_KERNEL', True),
+        ('cflags', 'KCFLAGS', True),
         ('override-host-type', 'OVERRIDE_HOST_TYPE', True),
+        ('cross-compile-compat', 'CROSS_COMPILE_COMPAT', True),
     )
 
     flavour_makeflags_build = (
@@ -361,6 +367,7 @@ class Gencontrol(Base):
                                                flavour)
         config_image = self.config.merge('image', arch, featureset, flavour)
 
+        vars['flavour'] = vars['localversion'][1:]
         vars['class'] = config_description['hardware']
         vars['longclass'] = (config_description.get('hardware-long')
                              or vars['class'])
@@ -385,6 +392,7 @@ class Gencontrol(Base):
         headers = self.templates["control.headers"]
         assert len(headers) == 1
 
+        do_meta = self.config.merge('packages').get('meta', True)
         config_entry_base = self.config.merge('base', arch, featureset,
                                               flavour)
         config_entry_build = self.config.merge('build', arch, featureset,
@@ -471,10 +479,7 @@ class Gencontrol(Base):
 
         packages_own = []
 
-        if not self.disable_signed:
-            build_signed = config_entry_build.get('signed-code')
-        else:
-            build_signed = False
+        build_signed = config_entry_build.get('signed-code')
 
         image = self.templates[build_signed and "control.image-unsigned"
                                or "control.image"]
@@ -486,25 +491,55 @@ class Gencontrol(Base):
         packages_own.append(image_main)
         makeflags['IMAGE_PACKAGE_NAME'] = image_main['Package']
 
+        # The image meta-packages will depend on signed linux-image
+        # packages where applicable, so should be built from the
+        # signed source packages
+        if do_meta and not build_signed:
+            packages_own.extend(self.process_packages(
+                self.templates["control.image.meta"], vars))
+            self.substitute_debhelper_config(
+                "image.meta", vars,
+                "linux-image%(localversion)s" % vars)
+
         package_headers = self.process_package(headers[0], vars)
         package_headers['Depends'].extend(relations_compiler_headers)
         packages_own.append(package_headers)
         if extra.get('headers_arch_depends'):
             extra['headers_arch_depends'].append('%s (= ${binary:Version})' %
                                                  packages_own[-1]['Package'])
+        if do_meta:
+            packages_own.extend(self.process_packages(
+                self.templates["control.headers.meta"], vars))
+            self.substitute_debhelper_config(
+                'headers.meta', vars,
+                'linux-headers%(localversion)s' % vars)
 
         if config_entry_build.get('vdso', False):
             makeflags['VDSO'] = True
 
-        if not self.disable_debug:
-            build_debug = config_entry_build.get('debug-info')
-        else:
-            build_debug = False
+        build_debug = config_entry_build.get('debug-info')
+
+        if os.getenv('DEBIAN_KERNEL_DISABLE_DEBUG'):
+            if self.changelog[0].distribution == 'UNRELEASED':
+                import warnings
+                warnings.warn('Disable debug infos on request '
+                              '(DEBIAN_KERNEL_DISABLE_DEBUG set)')
+                build_debug = False
+            else:
+                raise RuntimeError(
+                    'Unable to disable debug infos in release build '
+                    '(DEBIAN_KERNEL_DISABLE_DEBUG set)')
 
         if build_debug:
             makeflags['DEBUG'] = True
             packages_own.extend(self.process_packages(
                 self.templates['control.image-dbg'], vars))
+            if do_meta:
+                packages_own.extend(self.process_packages(
+                    self.templates["control.image-dbg.meta"], vars))
+                self.substitute_debhelper_config(
+                    'image-dbg.meta', vars,
+                    'linux-image%(localversion)s-dbg' % vars)
 
         merge_packages(packages, packages_own, arch)
 
@@ -573,11 +608,10 @@ class Gencontrol(Base):
         if build_debug:
             makeflags['KCONFIG_OPTIONS'] += ' -o DEBUG_INFO=y'
         if build_signed:
-            makeflags['KCONFIG_OPTIONS'] += ' -o LOCK_DOWN_IN_EFI_SECURE_BOOT=y -o MODULE_SIG=y'
+            makeflags['KCONFIG_OPTIONS'] += ' -o MODULE_SIG=y'
         # Add "salt" to fix #872263
-        makeflags['KCONFIG_OPTIONS'] += (' -o "BUILD_SALT=\\"%s%s\\""' %
-                                         (vars['abiname'],
-                                          vars['localversion']))
+        makeflags['KCONFIG_OPTIONS'] += \
+            ' -o "BUILD_SALT=\\"%(abiname)s%(localversion)s\\""' % vars
 
         cmds_binary_arch = ["$(MAKE) -f debian/rules.real binary-arch-flavour "
                             "%s" %
@@ -599,22 +633,14 @@ class Gencontrol(Base):
                      cmds=["$(MAKE) -f debian/rules.real %s %s" %
                            (merged_config, makeflags)])
 
-        # Substitute kernel version etc. into maintainer scripts,
-        # translations and lintian overrides
-        self._substitute_file('headers.postinst', vars,
-                              'debian/linux-headers-%s%s.postinst' %
-                              (vars['abiname'], vars['localversion']))
-        for name in ['postinst', 'postrm', 'preinst', 'prerm']:
-            self._substitute_file('image.%s' % name, vars,
-                                  'debian/%s.%s' %
-                                  (image_main['Package'], name))
+        self.substitute_debhelper_config(
+            'headers', vars,
+            'linux-headers-%(abiname)s%(localversion)s' % vars)
+        self.substitute_debhelper_config('image', vars, image_main['Package'])
         if build_debug:
-            debug_lintian_over = (
-                'debian/linux-image-%s%s-dbg.lintian-overrides' %
-                (vars['abiname'], vars['localversion']))
-            self._substitute_file('image-dbg.lintian-overrides', vars,
-                                  debug_lintian_over)
-            os.chmod(debug_lintian_over, 0o755)
+            self.substitute_debhelper_config(
+                'image-dbg', vars,
+                'linux-image-%(abiname)s%(localversion)s-dbg' % vars)
 
     def process_changelog(self):
         version = self.version = self.changelog[0].version
@@ -635,6 +661,8 @@ class Gencontrol(Base):
             'source_package': self.changelog[0].source,
             'abiname': self.abiname_version + self.abiname_part,
         }
+        self.vars['source_suffix'] = \
+            self.changelog[0].source[len(self.vars['source_basename']):]
         self.config['version', ] = {'source': self.version.complete,
                                     'upstream': self.version.linux_upstream,
                                     'abiname_base': self.abiname_version,
